############################################################################################################################
#									
# Routine :	alm_integrator.pl	v1.0.0.1			
#   									
# Description :								
#	This perl script implements ALM Agile Automation functionality. The current policies supported are:
#   Accept_ALMRequests
#	Create_ALMTasks
# 	Activate_ALMTasks
# 	Complete_ALMTasks
# 	Create_ALMActivities
#   Open_ALMActivities
#	wfa_Relate_Activities
#	wfa_Relate_Tasks
#   wfa_Sequence_Activities
#   wfa_Sequence_Tasks
#   
#   See ALM documentation for additional details.
#									
# Syntax :								
#	ratlperl alm_integrator.pl {-user <user>} {-pw <pw>} {-dbname <db>} {-dbset <dbset>}  [-CreateTask]  [-CreateActivity]
#   example: ratlperl alm_integrator.pl -user admin -pw "" -dbname CQALM -dbset CQMS.ALM.GALWAY
#																				
#									
# Licensed Materials - Property of IBM
# Rational ClearQuest
# (C) Copyright IBM Corp. 2009-2014.  All Rights Reserved.
# US Government Users Restricted Rights -
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp
######################################################################################################################
#
# Change Log:
#
# Ver       CRs                 Date         Programmer        Comments (fixed, fixed with dependency, added feature)
# 	 	    	                10/09        majidi 		   Created the script
#
# Known issues:
# 
######################################################################################################################
#!ratlperl -l -w
use Config;
use strict;
use Getopt::Long;
use File::Basename;
use CQPerlExt;
my $result;
my $userLogin ; 
my %opts;
my $session;
my $localreplica; 
my $VERSION="1.0.0.1";
use subs qw(main Policy_Accept_ALMRequests Policy_Create_ALMTasks Policy_Activate_ALMTasks Policy_Complete_ALMTasks Policy_Create_ALMActivities Policy_Open_ALMActivities SetEntityFieldVals ALM_SetFieldValue AddEntityFieldVals ALM_AddFieldValue CQValidateCommit Usage wfa_Relate_Activities wfa_Relate_Tasks wfa_Sequence_Activities);
# Return 1/exit 0 -> no error
# Unbuffer STDERR and STDOUT
select STDERR;  
$| = 1;
select STDOUT;
$| = 1;
#Define Logger parameters
use Logger qw(CONSOLE_ON CONSOLE_OFF CONSOLE_TERSE);
#use Logger qw(CONSOLE_ON);
my $logger = Logger->new(); # the logger element
# Create the utility object
use UtilLib;
my $ut = UtilLib->new();
my $log_time = $ut->ut_file_name_date_time_stamp();
my $progname = basename $0;
my $result = main();
if ($result) {
    exit(0); 
} else {
    exit(1); 
}
#######################################################################################################################
#sub main - Obtains the opts, sets the session variable, and calls the policy functions
#######################################################################################################################
sub main {
	my $failure;
	$logger->log_print_message(" main() ͂Ă܂B"); 
	# Define the input parameters
	my @opts = qw(
    	user=s		
	    pw=s		
	    dbname=s		
	    dbset=s		
	    logfile:s
	    CreateTask
	    CreateActivity
	    preview	
	    help
	);

	my %opts = (
	    user => undef,
	    pw => undef,
	    dbname => undef,
	    dbset => undef,
	    logfile => undef,
	);
	my @required_opts = qw ( user pw dbname dbset);
	
	# Get the input parameters
	unless (GetOptions(\%opts, @opts)) {
	 	$logger->log_print_message("R}hsւ̈p@:");	
	     my $retval = &Usage;
	     return 0;
	
	}
	
	#Define the LogFile
	if ($opts{logfile}) {
		my $retval;
		my $logdir = dirname($opts{logfile});
		if ( ! -e $logdir) {
			$retval = $ut->ut_create_directory_tree($logdir);
			if ( ! $retval) {
	          		$logger->log_print_error(__LINE__ . "O t@C fBNg $logdir 쐬ł܂ł");    	
	          		$logger->log_print_error(__LINE__ . "XNvg̎śAȂŌp܂B");
	       	}
		}
		#$retval = $logger->log_set_file_on($opts{logfile});
		$retval = $logger->log_prv_open_file($opts{logfile});
		if ( ! $retval) {
			$logger->log_print_error(__LINE__ . "O t@C $opts{logfile} J܂łB");
			$logger->log_print_error(__LINE__ . "XNvg̎śAȂŌp܂B");
	 	 }
		$logger->log_print_message("MO $log_time ɊJn܂B"); 
	}	 
	 
	# Display the help usage
	if ($opts{help}) {
		&Usage();
		return 1;
	}
	
	# Display preview inputs to log file 
	if ($opts{preview}){
		$logger->log_print_message("`ꂽ[U[:");
		$logger->log_print_message("`ꂽ[U[ OC $opts{user} łB");
	    $logger->log_print_message("`ꂽf[^x[X $opts{dbname} łB");
		$logger->log_print_message("`ꂽ dbset  $opts{dbset} łB");
		$logger->log_print_message("`ꂽO t@C $opts{logfile} łB");
		return 1;
	}
	# End program if any of the required parameters are not defined
	my  $message = "";
	for my $req (@required_opts )  {
			$message .= "K{wIvV -$req  ̌\n" if !defined $opts{$req};
			if ($message ) {
				$logger->log_print_message("\nKvȃR}hsIvVĂ܂AR}hsւ̈p@:");
				$logger->log_print_message("$message");
				my $retval = &Usage();
				return 0;
		}
	}
	
	# Check to see if an instance of the program is already running, if so, print an error message and exit
	use Fcntl qw(:flock);
	my $lockfile = $0.'lock_file';
	my $fhpid;
	if (!open($fhpid, '>', $lockfile)) {
		$logger->log_print_message("bN t@C '".$lockfile."' JƂł܂B$!");
	}
	if (!flock($fhpid, LOCK_EX|LOCK_NB)) {
		$logger->log_print_message("'".$0."' ̃CX^XɎsłBI܂B");
		return 0;
	}
	
	# Start a CQ session
	$session = CQPerlExt::CQSession_Build();
	# Logon to the specified database
	eval {$session->UserLogon("$opts{user}", "$opts{pw}", "$opts{dbname}", "$opts{dbset}")};
	if ($@) {
		$logger->log_print_error(__LINE__ . "$opts{dbname} ւ̃OCɎs܂: $@"); 
		return 0;
	}
	$userLogin = $session->GetUserLoginName();
	$localreplica = ($session->GetLocalReplica())->GetDisplayName(); 
	$logger->log_print_success("$userLogin ɂ $opts{dbname} ւ̃OC܂B"); 
	
	if ($opts{CreateTask}) {
		$failure=Policy_Create_ALMTasks();
		if ($failure == 0){
			CQPerlExt::CQSession_Unbuild($session);
			return 0;
		}
	}
	
	$failure=Policy_Complete_ALMTasks();
	if ($failure == 0){
		CQPerlExt::CQSession_Unbuild($session);
		return 0;
	}
	
	if ($opts{CreateActivity}) {
		$failure=Policy_Create_ALMActivities();
		if ($failure == 0){
			CQPerlExt::CQSession_Unbuild($session);
			return 0;
		}
	}
		
	$failure=Policy_Activate_ALMTasks();
	if ($failure == 0){
		CQPerlExt::CQSession_Unbuild($session);
		return 0;
	}
	
	$failure=Policy_Open_ALMActivities();
	if ($failure == 0){
		CQPerlExt::CQSession_Unbuild($session);
		return 0;
	}
	
	$failure=Policy_Accept_ALMRequests();
	if ($failure == 0){
		CQPerlExt::CQSession_Unbuild($session);
		return 0;
	}
	
	$failure=wfa_Relate_Activities();
	if ($failure == 0){
		CQPerlExt::CQSession_Unbuild($session);
		return 0;
	}
	
	$failure=wfa_Relate_Tasks();
	if ($failure == 0){
		CQPerlExt::CQSession_Unbuild($session);
		return 0;
	}
	$failure=wfa_Sequence_Activities();
	if ($failure == 0){
		CQPerlExt::CQSession_Unbuild($session);
		return 0;
	}
	
	$failure=wfa_Sequence_Tasks();
	if ($failure == 0){
		CQPerlExt::CQSession_Unbuild($session);
		return 0;
	}
	
	CQPerlExt::CQSession_Unbuild($session);
	$logger->log_print_message(" main() I܂B"); 
	return 1;
}
###################################################################################################################
# sub Policy_Accept_ALMRequests - Accept an ALMRequest
###################################################################################################################
sub Policy_Accept_ALMRequests { 
	
	my $querydef;
	my $operator;
	my $resultset;
	my $count;
	my $querydef2;
	my $operator2;
	my $resultset2;
	my $count2;
	my $operator3;
	my $operator4;
	my @active_open=("Activated","Opened");
	my $res_codes;
	my $ALMRequest;
	my %ARH=();
	my %attributes;
	my $failure;
	my $AREntity;
	my $tasks;
	my $TaskEntity;
	my $TotalHours;
	my $currEntity;


	$logger->log_print_message("Policy_Accept_ALMRequests() ͂Ă܂\n");

	# Query ALMRequests that have at least one completed task
	$querydef = $session->BuildQuery("ALMRequest");
	$querydef->BuildField("ID");
	$operator = $querydef->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
	$operator->BuildFilter("State", $CQPerlExt::CQ_COMP_OP_EQ, ["Opened"]);
	$operator->BuildFilter("Project.Obsolete", $CQPerlExt::CQ_COMP_OP_NEQ,["Yes"]);
	$operator->BuildFilter("Project.RequestPolicy_Accept", $CQPerlExt::CQ_COMP_OP_EQ,["Yes"]);
	$operator->BuildFilter("Tasks.State", $CQPerlExt::CQ_COMP_OP_EQ, ["Completed"]);
	$operator->BuildFilter("ratl_mastership", $CQPerlExt::CQ_COMP_OP_EQ, ["$localreplica"]);
	$resultset = $session->BuildResultSet($querydef);
	$resultset->EnableRecordCount();			
	$resultset->Execute;	
	$count = $resultset->GetRecordCount();	
	# Query ALMTasks that reference each ALMRequest from above and are 
	# (Activated and Opened) or (Completed and ResolutionCode outside the policy)
	while ($resultset->MoveNext eq $CQPerlExt::CQ_SUCCESS) {
		$ALMRequest=$resultset->GetColumnValue(1);		
		$AREntity = $session->GetEntity("ALMRequest", $ALMRequest);
		$res_codes=$AREntity->GetFieldStringValueAsList("Project.RequestPolicy_ResCodes");
		
		$querydef2 = $session->BuildQuery("ALMTask");
		$querydef2->BuildField("ID");
		$operator2 = $querydef2->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_OR);
		$operator3 = $operator2->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
		$operator4 = $operator2->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);

		if ($$res_codes[0] ne ""){
			$operator3->BuildFilter("Request", $CQPerlExt::CQ_COMP_OP_EQ, ["$ALMRequest"]);
			$operator3->BuildFilter("ResolutionCode", $CQPerlExt::CQ_COMP_OP_NOT_IN,\@$res_codes);
			$operator3->BuildFilter("State", $CQPerlExt::CQ_COMP_OP_NOT_IN,\@active_open);
		}
		$operator4->BuildFilter("Request", $CQPerlExt::CQ_COMP_OP_EQ, ["$ALMRequest"]);
		$operator4->BuildFilter("State", $CQPerlExt::CQ_COMP_OP_IN,\@active_open);
		
		$resultset2 = $session->BuildResultSet($querydef2);
		$resultset2->EnableRecordCount();			
		$resultset2->Execute;	
		$count2 = $resultset2->GetRecordCount();	
		if ($count2){
		   next;		
		}
		else{
			# Store the value of the ALMRequests for later processing
       	     $ARH{$ALMRequest}=1;
		}
		
	}
	# Accept each request
	foreach my $Request ( keys %ARH ) {
		$currEntity = $session->GetEntity("ALMRequest", $Request);
		$tasks = $currEntity->GetFieldStringValueAsList("Tasks");
 		foreach my $task (@$tasks) 
		{
			$TaskEntity=$session->GetEntity("ALMTask", "$task");
			$TotalHours= $TotalHours + $TaskEntity->GetFieldStringValue("ActualEffort");
		
		}
		$logger->log_print_message("Ԃ̍v $TotalHours ł\n");
		if ($TotalHours != 0){
			$attributes{"ActualEffort"}=$TotalHours;
		}
		if (SetEntityFieldVals("ALMRequest",$Request,"Accept",%attributes) == 0){
			return 0;
		}
		else{
			$logger->log_print_message("ALMRequest $Request 𐳏Ɏ󂯓܂\n");
		}
	}
	$logger->log_print_message("Policy_Accept_ALMRequests() I܂\n");	
	return 1;
}
###################################################################################################################
# sub Policy_Create_ALMTasks - Call CreateTask to create Tasks for Open ALMRequests
###################################################################################################################
sub Policy_Create_ALMTasks { 
	
	my $querydef;
	my $operator;
	my $resultset;
	my $count;
	my $ALMRequest;
	my %ARH=();
	my $failure;
	my $currEntity;
	

	$logger->log_print_message("Policy_Create_ALMTasks() ͂܂\n");

	# Query ALMRequests that have no tasks
	$querydef = $session->BuildQuery("ALMRequest");
	$querydef->BuildField("ID");
	$operator = $querydef->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
	$operator->BuildFilter("State", $CQPerlExt::CQ_COMP_OP_EQ, ["Opened"]);
	$operator->BuildFilter("Project.Obsolete", $CQPerlExt::CQ_COMP_OP_NEQ,["Yes"]);
	$operator->BuildFilter("Project.TaskPolicy_Create", $CQPerlExt::CQ_COMP_OP_EQ,["Yes"]);
	$operator->BuildFilter("Tasks", $CQPerlExt::CQ_COMP_OP_IS_NULL, [""]);
	$resultset = $session->BuildResultSet($querydef);
	$resultset->EnableRecordCount();			
	$resultset->Execute;	
	$count = $resultset->GetRecordCount();	
	while ($resultset->MoveNext eq $CQPerlExt::CQ_SUCCESS) {
		$ALMRequest=$resultset->GetColumnValue(1);		
        $ARH{$ALMRequest}=1;
	}
	# For each request create tasks as configured by WorkConfiguraiton record
	foreach my $Request ( keys %ARH ) {
		my $currEntity = $session->GetEntity("ALMRequest", "$Request");
		$logger->log_print_message("ALMRequest $Request ̃^XN쐬 CreateTask ĂяoĂ܂ \n");
		$failure=$currEntity->FireNamedHook("CreateTask","");
		if ($failure){
			$logger->log_print_error(__LINE__ ." ALMRequest $Request  CreateTask ̃G[\n");
			return 0;	
		}
		else{
			$logger->log_print_message("ALMRequest $Request  CreateTask 𐳏ɌĂяo܂\n"); 
		}
	}
	$logger->log_print_message("Policy_Create_ALMTasks() I܂\n");	
	return 1;
}
###################################################################################################################
# sub Policy_Activate_ALMTasks - Activates ALMTasks when primary activities are created
###################################################################################################################
sub Policy_Activate_ALMTasks { 
	
	my $querydef;
	my $operator;
	my $resultset;
	my $count;
	my $ALMTask;
	my %ATH=();
	my $failure;
	my $currEntity;
	my $Owner;
	my $roleRecs;
	my $roleEntity;
	my $primaryOwner;
	my %attributes;
	my $Priority;
	my $ALMType;
	

	$logger->log_print_message("Policy_Activate_ALMTasks() ͂Ă܂\n");

	# Query ALMTasks that have activities
	$querydef = $session->BuildQuery("ALMTask");
	$querydef->BuildField("ID");
	$operator = $querydef->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
	$operator->BuildFilter("State", $CQPerlExt::CQ_COMP_OP_EQ, ["Opened"]);
	$operator->BuildFilter("Project.Obsolete", $CQPerlExt::CQ_COMP_OP_NEQ,["Yes"]);
	$operator->BuildFilter("Project.TaskPolicy_Activate", $CQPerlExt::CQ_COMP_OP_EQ,["Yes"]);
	$operator->BuildFilter("Activities", $CQPerlExt::CQ_COMP_OP_IS_NOT_NULL, [""]);
	$operator->BuildFilter("ratl_mastership", $CQPerlExt::CQ_COMP_OP_EQ, ["$localreplica"]);
	$resultset = $session->BuildResultSet($querydef);
	$resultset->EnableRecordCount();			
	$resultset->Execute;	
	$count = $resultset->GetRecordCount();	
	while ($resultset->MoveNext eq $CQPerlExt::CQ_SUCCESS) {
		$ALMTask=$resultset->GetColumnValue(1);		
        $ATH{$ALMTask}=1;
	}
	# For each task set the Owner to Primary contact and set the Priority to Unassigned
 	foreach my $Task ( keys %ATH ) {
	 	$Owner="";
	 	%attributes=();
 		$currEntity = $session->GetEntity("ALMTask", $Task);
 		$primaryOwner = $currEntity->GetFieldStringValue("PrimaryRole.Primary");
 		$roleRecs = $currEntity->GetFieldStringValueAsList("Roles");
 		if ($primaryOwner){
 			$Owner=$primaryOwner;		
 		}
 		elsif ($$roleRecs[0] ne ""){
 			@{$roleRecs}= sort(@{$roleRecs});
     	    $roleEntity = $session->GetEntity("ALMRole",$$roleRecs[0]);
 			$Owner= $roleEntity->GetFieldStringValue("Primary");
 		}
 		else{
 			$logger->log_print_error(__LINE__ ." ALMTask:$Task ̓KL҂邱Ƃł܂ł");
 			return 0;	
 		}
 		%attributes = (
 			Owner => $Owner,
 		);
 		eval{$session->GetEntity("ALMType","ALMTask ɂ͗Dx蓖ĂĂ܂")};
 		if ($@){
	 		$logger->log_print_warning("Dx ALMTask:$Task ɑ΂Ė蓖Ăɐݒł܂BALMType ̖蓖Ă ASAP 쐬܂B\n");	
	 	}
	 	else{
			$attributes{"Priority"}="蓖";
		}

 		if (SetEntityFieldVals("ALMTask",$Task,"Activate",%attributes) == 0){
	 		return 0;	
 		}
 		else{
	 		$logger->log_print_message("ALMTask:$Task 𐳏ɃANeBu܂\n");
 		}
 	}
 	$logger->log_print_message("Policy_Activate_ALMTasks() I܂\n");	
	return 1;
}
###################################################################################################################
# sub Policy_Complete_ALMTasks - Complete Tasks with Resolution Code specified when all 
# Activities are completed with one of the Resolution Codes listed
###################################################################################################################
sub Policy_Complete_ALMTasks { 
	
	my $querydef;
	my $operator;
	my $resultset;
	my $count;
	my $querydef2;
	my $operator2;
	my $resultset2;
	my $count2;
	my $operator3;
	my $operator4;
	my @active_open=("Activated","Opened");
	my @submit_open_active=("Activated","Opened","Submitted");
	my $res_codes;
	my $res_code;
	my $ALMRequest;
	my %ATH=();
	my %attributes;
	my $failure;
	my $ATEntity;
	my $ALMTask;
	my $currEntity;
	my $res_code;
	my $ActualEffort;
	my $activities;
	my $ActEntity;
	my $TotalHours;


	$logger->log_print_message("Policy_Complete_ALMTasks() ͂Ă܂\n");

	# Query ALMTasks that have at least one completed activity
	$querydef = $session->BuildQuery("ALMTask");
	$querydef->BuildField("ID");
	$operator = $querydef->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
	$operator->BuildFilter("State", $CQPerlExt::CQ_COMP_OP_EQ,\@active_open);
	$operator->BuildFilter("Project.Obsolete", $CQPerlExt::CQ_COMP_OP_NEQ,["Yes"]);
	$operator->BuildFilter("Project.TaskPolicy_Complete", $CQPerlExt::CQ_COMP_OP_EQ,["Yes"]);
	$operator->BuildFilter("Activities.State", $CQPerlExt::CQ_COMP_OP_EQ, ["Completed"]);
	$operator->BuildFilter("ratl_mastership", $CQPerlExt::CQ_COMP_OP_EQ, ["$localreplica"]);
	$resultset = $session->BuildResultSet($querydef);
	$resultset->EnableRecordCount();			
	$resultset->Execute;	
	$count = $resultset->GetRecordCount();	
 	# Query ALMActivities that reference each ALMTask from above and are 
 	# (Activated, Opened, or Submitted) or (Completed and ResolutionCode outside the policy)
	while ($resultset->MoveNext eq $CQPerlExt::CQ_SUCCESS) {
		$ALMTask=$resultset->GetColumnValue(1);		
		$ATEntity = $session->GetEntity("ALMTask", $ALMTask);
		$res_codes=$ATEntity->GetFieldStringValueAsList("Project.TaskPolicy_ResCodes");
		$querydef2 = $session->BuildQuery("ALMActivity");
		$querydef2->BuildField("ID");
		$operator2 = $querydef2->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_OR);
		$operator3 = $operator2->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
		$operator4 = $operator2->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);

		if ($$res_codes[0] ne ""){
			$operator3->BuildFilter("Task", $CQPerlExt::CQ_COMP_OP_EQ, ["$ALMTask"]);
			$operator3->BuildFilter("ResolutionCode", $CQPerlExt::CQ_COMP_OP_NOT_IN,\@$res_codes);
			$operator3->BuildFilter("State", $CQPerlExt::CQ_COMP_OP_NOT_IN,\@submit_open_active);
		}
		$operator4->BuildFilter("Task", $CQPerlExt::CQ_COMP_OP_EQ, ["$ALMTask"]);
		$operator4->BuildFilter("State", $CQPerlExt::CQ_COMP_OP_IN,\@submit_open_active);
		
		$resultset2 = $session->BuildResultSet($querydef2);
		$resultset2->EnableRecordCount();			
		$resultset2->Execute;	
		$count2 = $resultset2->GetRecordCount();	
		if ($count2){
		   next;		
		}
		else{
			# Store the value of the ALMTasks for later processing
       	     $ATH{$ALMTask}=1;
		}
 	}
 	# Complete each task
	foreach my $Task ( keys %ATH ) {
		$res_code="";
		$ActualEffort=0;
		%attributes=();
		$currEntity = $session->GetEntity("ALMTask", $Task);
 		$res_code = $currEntity->GetFieldStringValue("Project.TaskPolicy_ResCode");
  		if ($res_code){
 			%attributes = (
 				ResolutionSummary => $res_code,
 				ResolutionCode => $res_code,
 			);
 		}
 		else{
 			$logger->log_print_error(__LINE__ ." ALMTask:$Task ̓KR[h邱Ƃł܂ł");
 			return 0;	
 		}
 		$activities = $currEntity->GetFieldStringValueAsList("Activities");
 		foreach my $activity (@$activities) 
		{
			$ActEntity=$session->GetEntity("ALMActivity", "$activity");
			$TotalHours= $TotalHours + $ActEntity->GetFieldStringValue("ActualEffort");
		
		}
		$logger->log_print_message("Ԃ̍v $TotalHours ł\n");
		if ($TotalHours != 0){
			$attributes{"ActualEffort"}=$TotalHours;
		}
 		if (SetEntityFieldVals("ALMTask",$Task,"Complete",%attributes) == 0){
	 		return 0;	
 		}
 		else{
	 		$logger->log_print_message("ALMTask:$Task 𐳏Ɋ܂\n");
 		}
	}
	$logger->log_print_message("Policy_Complete_ALMTasks() I܂\n");	
 	return 1;
}
###################################################################################################################
# sub Policy_Create_Activities - Call CreateActivity to create Activities for Open ALMTasks
###################################################################################################################
sub Policy_Create_ALMActivities { 
	
	my $querydef;
	my $operator;
	my $resultset;
	my $count;
	my $ALMTask;
	my %ATH=();
	my $failure;
	my $currEntity;
	my @active_open=("Activated","Opened");
	

	$logger->log_print_message("Policy_Create_ALMActivities() ͂Ă܂\n");

	# Query ALMTasks that have no activities
	$querydef = $session->BuildQuery("ALMTask");
	$querydef->BuildField("ID");
	$operator = $querydef->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
	$operator->BuildFilter("State", $CQPerlExt::CQ_COMP_OP_EQ, \@active_open);
	$operator->BuildFilter("Project.Obsolete", $CQPerlExt::CQ_COMP_OP_NEQ,["Yes"]);
	$operator->BuildFilter("Project.ActivityPolicy_Create", $CQPerlExt::CQ_COMP_OP_EQ,["Yes"]);
	$operator->BuildFilter("Activities", $CQPerlExt::CQ_COMP_OP_IS_NULL, [""]);
	$operator->BuildFilter("Type", $CQPerlExt::CQ_COMP_OP_IS_NOT_NULL, [""]);
	$resultset = $session->BuildResultSet($querydef);
	$resultset->EnableRecordCount();			
	$resultset->Execute;	
	$count = $resultset->GetRecordCount();	
	while ($resultset->MoveNext eq $CQPerlExt::CQ_SUCCESS) {
		$ALMTask=$resultset->GetColumnValue(1);		
        $ATH{$ALMTask}=1;
	}
	# Call CreateActivity record script alias to create the activities
	foreach my $Task ( keys %ATH ) {
		my $currEntity = $session->GetEntity("ALMTask", "$Task");
		$logger->log_print_message("CreateActivity ĂяoāAALMTask:$Task ̃ANeBreB쐬܂B\n");
		$failure=$currEntity->FireNamedHook("CreateActivity","");
		if ($failure) {
			$logger->log_print_error(__LINE__ ." ALMTask:$Task  CreateActivity ̃G[\n");
			return 0;
		}
		else{
			$logger->log_print_message("ALMTask:$Task 𐳏ɌĂяo܂\n"); 
		}
	}
	$logger->log_print_message("Policy_Create_ALMActivities() I܂\n");	
	return 1;
}
###################################################################################################################
# sub Policy_Open_Activities - Open Activities and set owner
###################################################################################################################
sub Policy_Open_ALMActivities { 
	
	my ($Task)= @_;
	my $querydef;
	my $operator;
	my $resultset;
	my $count;
	my $ALMActivity;
	my %AAH=();
	my $failure;
	my $currEntity;
	my $failure;
	my $Owner;
	my $roleRecs;
	my $roleEntity;
	my %attributes;
	my $OwnerFlagActivityPolicy;
	my $roleOwner;
	my $TaskOwner;
	my $members;
	my $groups;
	my $querydef;
	my $operator;
	my $operator2;
	my $resultset;
	my %OwnersList;
	my $primaryOwner;


	$logger->log_print_message("Policy_Open_ALMActivities() ͂Ă܂\n");
	
	# Query ALMActivities in submitted state
	$querydef = $session->BuildQuery("ALMActivity");
	$querydef->BuildField("ID");
	$operator = $querydef->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
	$operator->BuildFilter("State", $CQPerlExt::CQ_COMP_OP_EQ, ["Submitted"]);
	$operator->BuildFilter("Project.Obsolete", $CQPerlExt::CQ_COMP_OP_NEQ,["Yes"]);
	$operator->BuildFilter("Project.ActivityPolicy_Create", $CQPerlExt::CQ_COMP_OP_EQ,["Yes"]);
	$operator->BuildFilter("ratl_mastership", $CQPerlExt::CQ_COMP_OP_EQ, ["$localreplica"]);
	$resultset = $session->BuildResultSet($querydef);
	$resultset->EnableRecordCount();			
	$resultset->Execute;	
	$count = $resultset->GetRecordCount();	

	while ($resultset->MoveNext eq $CQPerlExt::CQ_SUCCESS) {
		$ALMActivity=$resultset->GetColumnValue(1);		
        $AAH{$ALMActivity}=1;
	}
	# For each activity set the owner and move to open
 	foreach my $Activity ( keys %AAH ) {
	 	$Owner="";
	 	%OwnersList=();
	 	%attributes=();
  		$currEntity = $session->GetEntity("ALMActivity", $Activity);
  		$OwnerFlagActivityPolicy=$currEntity->GetFieldStringValue("Project.ActivityPolicy_OwnerFlag");
   		$roleRecs = $currEntity->GetFieldStringValueAsList("Roles");

		foreach my $role (@$roleRecs) 
		{
			$roleEntity = $session->GetEntity("ALMRole","$role");
			next if ($roleEntity->GetFieldStringValue("Obsolete") =~ /Yes/i);
	
			$members = $roleEntity->GetFieldStringValueAsList("Members");
			$groups = $roleEntity->GetFieldStringValueAsList("Groups");
	
			$querydef = $session->BuildQuery("users");	
			$querydef->BuildField("login_name");	
			$operator = $querydef->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
			$operator->BuildFilter("is_active", $CQPerlExt::CQ_COMP_OP_EQ,[1]);
	
			$operator2 = $operator->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_OR); 
			$operator2->BuildFilter("login_name",$CQPerlExt::CQ_COMP_OP_EQ, \@$members);
			$operator2->BuildFilter("groups", $CQPerlExt::CQ_COMP_OP_IN, \@$groups) unless (grep(/^Everyone$/, @{$groups}));
	
			$resultset = $session->BuildResultSet($querydef);
			$resultset->Execute;
			while ($resultset->MoveNext eq $CQPerlExt::CQ_SUCCESS)
			{
				$OwnersList{$resultset->GetColumnValue(1)} = 1;
			}
		}
  		if ($OwnerFlagActivityPolicy eq "TaskOwner"){
	  		$TaskOwner=$currEntity->GetFieldStringValue("Task.Owner");
	  		if ($OwnersList{"$TaskOwner"}){
		  		$Owner=$TaskOwner;
		  		$logger->log_print_message("L҂ ALMActivity:$Activity  $Owner ɐݒ肳܂\n");
	  			%attributes = (
					Owner => $Owner,
	 			);
	  		}
	  		else{
		 		$logger->log_print_warning("L $TaskOwner  ALMActivity:$Activity ̎󂯓\ȏL҂̃Xgɂ܂BASAP ܂B\n");		 		
 			}
  		}
  		if ($OwnerFlagActivityPolicy  eq "Unassigned"){
 			if ($OwnersList{"蓖"}){
	 			$Owner="蓖";
		 		$logger->log_print_message("L҂ ALMActivity:$Activity  $Owner ɐݒ肳܂\n");
				%attributes = (
					Owner => $Owner,
	 			);	
	 		}
	 		else{
		 		$logger->log_print_warning("L҂ ALMActivity:$Activity ɑ΂Ė蓖Ăɐݒł܂B ASAP ɖ蓖Ăȃ[U[ǉ܂B\n");		 		
 			}
  		}
  		if (($OwnerFlagActivityPolicy eq "PrimaryContact") || ($Owner eq "")){
	 		$primaryOwner = $currEntity->GetFieldStringValue("PrimaryRole.Primary");
	 		$roleRecs = $currEntity->GetFieldStringValueAsList("Roles");
	 		if ($primaryOwner){
				$Owner = $primaryOwner;
	 		}
	 		elsif ($$roleRecs[0] ne ""){
				@{$roleRecs} = sort(@{$roleRecs});
				$roleEntity = $session->GetEntity("ALMRole",$$roleRecs[0]);
				$Owner = $roleEntity->GetFieldStringValue("Primary");
	 		}
	 		else{
				$logger->log_print_error(__LINE__ ." ALMTask:$Task ̓KL҂邱Ƃł܂ł");
				$Owner = "";
				return 0;
	 		}
	 		if ($Owner){
				$logger->log_print_message("L҂ ALMActivity:$Activity  $Owner ɐݒ肳܂\n");
				%attributes = (
					Owner => $Owner,
				);
	 		}
  		}
		if ($Owner eq ""){
			$logger->log_print_error(__LINE__ ." ALMActivity:$Activity ̓KL҂邱Ƃł܂ł");
			return 0;	
		}
	 	if (SetEntityFieldVals("ALMActivity",$Activity,"Open",%attributes) == 0){
			$logger->log_print_error(__LINE__ ." ALMTask:$Task ̃ANeBreB̃G[\n");
	 		return 0;	
 		}
 		else{
	 		$logger->log_print_message("ALMActivity:$Activity ɊJ܂\n");
 		}
 	}
 	$logger->log_print_message("Policy_Open_ALMActivities() I܂\n");	
	return 1;
}

#######################################################################################################################
# sub SetEntityFieldVals - Tranisiton to the target state, set values if any and commit
#######################################################################################################################
sub SetEntityFieldVals
{
	my ($status,$msg,$failure);
	my ($recordType,$id,$action,%attributes) = @_;
	my $failure;
	my $currFieldValue;

	
	$logger->log_print_message("SetEntityFieldVals() ͂Ă܂\n");
	my $currEntity = $session->GetEntity("$recordType", $id);
	$session->EditEntity($currEntity,"$action");
	foreach my $recFieldName ( keys %attributes ){
		$currFieldValue = $currEntity->GetFieldStringValue("$recFieldName");
		if ($currFieldValue eq ""){
			my $recFieldValue = $attributes{$recFieldName};
			$logger->log_print_message("ALM_SetFieldValue() ĂяoĂ܂\n");
			$failure = ALM_SetFieldValue($currEntity,"$recFieldName","$recFieldValue");
			if ($failure == 0){
				$currEntity->Revert();
				return 0;
			}
		}
	}
	if ($recordType eq "ALMTask"){
			$failure = $currEntity->SetFieldRequirednessForCurrentAction("Priority",$CQPerlExt::CQ_OPTIONAL);	
	}
   if (CQValidateCommit($currEntity) == 0) {
	$failure=0;
   }
   else{
    $failure=1;	
   }
   $logger->log_print_message("SetEntityFieldVals() I܂\n");
   return $failure;
}
#######################################################################################################################
#	sub ALM_SetFieldValue - Se to the field value and check for errors
#######################################################################################################################
sub ALM_SetFieldValue
{
	my ($entity,$fieldName,$fieldValue) = @_;
	my $failure;
	my $currEntityDef;
	my $id;
	$logger->log_print_message("ALM_SetFieldValue() ͂Ă܂\n");
	my $currEntityDef = $entity->GetEntityDefName();
	my $id=$entity->GetFieldStringValue("ID");
	$logger->log_print_message("$fieldName  $currEntityDef ̒l $fieldValue ɐݒ肵Ă܂:$id\n");
	$failure=$entity->SetFieldValue($fieldName,$fieldValue);
	if ($failure ne "") {
		$logger->log_print_error(__LINE__ ." $fieldName  $currEntityDef  $fieldValue ւ̐ݒɎs܂:$id\n");
		$logger->log_print_message("ALM_SetFieldValue() I܂\n");
		return 0;
	}
	$logger->log_print_message("ALM_SetFieldValue() I܂\n");
	return 1;
}
#######################################################################################################################
# sub AddEntityFieldVals - Tranisiton to the target state, Add values if any and commit
#######################################################################################################################
sub AddEntityFieldVals
{
	my ($status,$msg,$failure);
	my ($recordType,$id,$action,%attributes) = @_;
	my $failure;
	my $currFieldValue;
	
	$logger->log_print_message("AddEntityFieldVals() ͂Ă܂\n");
	my $currEntity = $session->GetEntity("$recordType", $id);
	$session->EditEntity($currEntity,"$action");
	foreach my $recFieldName ( keys %attributes ){
		$currFieldValue = $currEntity->GetFieldStringValue("$recFieldName");
		my $recFieldValue = $attributes{$recFieldName};
		$logger->log_print_message("ALM_AddFieldValue() ĂяoĂ܂\n");
		$failure = ALM_AddFieldValue($currEntity,"$recFieldName","$recFieldValue");
		if ($failure == 0){
			$currEntity->Revert();
			return 0;
		}
	}
    if (CQValidateCommit($currEntity) == 0) {
    	$failure=0;
    }
    else{
       $failure=1;	
   }
   $logger->log_print_message("SetEntityFieldVals() I܂\n");
   return $failure;
}
#######################################################################################################################
#	sub ALM_AddFieldValue - Adds the field value and check for errors
#######################################################################################################################
sub ALM_AddFieldValue
{
	my ($entity,$fieldName,$fieldValue) = @_;
	my $failure;
	my $currEntityDef;
	my $id;
	
	$logger->log_print_message("ALM_AddFieldValue() ͂Ă܂\n");
	my $currEntityDef = $entity->GetEntityDefName();
	my $id=$entity->GetFieldStringValue("ID");
	$logger->log_print_message("l $fieldValue  $currEntityDef ̃tB[h $fieldName ɒǉĂ܂:$id\n");
	$failure=$entity->AddFieldValue($fieldName,$fieldValue);
	if ($failure ne "") {
		$logger->log_print_error(__LINE__ ." $fieldName  $currEntityDef  $fieldValue ւ̒ǉɎs܂:$id\n");
		$logger->log_print_message("ALM_AddFieldValue() I܂\n");
		return 0;
	}
	$logger->log_print_message("ALM_AddFieldValue() I܂\n");
	return 1;
}
#######################################################################################################################
#	sub CQValidateCommit - Validate and Commit an entity
#######################################################################################################################
sub CQValidateCommit {
	
	my $Entity=shift(@_);
	my $status;
	my $msg;

	$logger->log_print_message("CQValidateCommit() ͂Ă܂\n");
	$status="";
	my $currEntityName = $Entity->GetDisplayName();
	my $currEntityType = $Entity->GetEntityDefName();
	eval {$status = $Entity->Validate()};

	if  ($@ || ($status ne "")) 
	{		
			$logger->log_print_message("$currEntityType: $currEntityName\n ɖ߂܂B\n");
			$Entity->Revert();
		    $msg = "؂Ɏs܂: $@\n";
		    $logger->log_print_error(__LINE__ ." $msg");
		    $msg = "߂Xe[^X $status łB\n";
			$logger->log_print_error(__LINE__ ." $msg");
			$logger->log_print_message("CQValidateCommit() I܂\n");
	    	return 0;
	}
	else
	{
		eval {$status = $Entity->Commit()};
		if ($@ || ($status ne ""))
		{	
			$logger->log_print_message("$currEntityType: $currEntityName\n ɖ߂܂B\n");
			$Entity->Revert();
			$msg = "R~bgɎs܂: $@\n";
		    $logger->log_print_error(__LINE__ ." $msg");
		    $msg = "߂Xe[^X $status łB\n";
			$logger->log_print_error(__LINE__ ." $msg");
			$logger->log_print_message("CQValidateCommit() I܂\n");
	       	return 0;
	
		}
		else {
			my $msg = "$currEntityType $currEntityName ̓R~bg܂B";
			$logger->log_print_success("$msg");
		}

	}
	$logger->log_print_message("CQValidateCommit() I܂\n");
	return 1;
}

#######################################################################################################################
# sub Usage - Display the usage
#######################################################################################################################
sub Usage {
my $message =
"gp@: $progname -user <username> -pw <password> -dbname <database name> -dbset <database set> [-CreateTask] [-CreateActivity] -logfile <logfile> [-preview] [-help]\n".
"\n".
"$progname $VERSION :ALM [Nt[p[eBeB\n".
"\n".
"-user <user>	   X[p[ [U[ ClearQuest [U[̃OC IDB\n".
"-pw <password>	   ClearQuest [U[̃pX[hB\n".
"-dbname <dbname>   f[^x[Xւ̃XL[} |WgڑB\n".
"-dbset <dbset>	   [U[ f[^x[XڑB\n".
"-CreateTask        ̃XNvgɂ ALM CreateTask Record XNvgʖs܂B}`TCgł̃|V[ŝ 1 ̃TCĝ݂łB\n".
"-CreateActivity    ̃XNvgɂ ALM CreateActivity Record XNvgʖs܂B}`TCgł̃|V[ŝ 1 ̃TCĝ݂łB\n".
"-logfile <logfile> R\[ɕ\ƋɁAw肳ꂽt@Cɏo͂OƂċL^܂B\n".
"-preview           O t@C̃[U[͂\܂B\n".
"-help              ̃bZ[Wo͂ďI܂B\n".
"\n";
print $message;
return 0;
}
###################################################################################################################
# sub wfa_Relate_Activities - Relate primary activities using the Related activities tab
###################################################################################################################
sub wfa_Relate_Activities { 
	
	my $querydef;
	my $operator;
	my $resultset;
	my @active_open=("Activated","Opened");
	my $count;
	my $ALMTask;
	my %ATH=();
	my $failure;
	my $currTask;
	my %attributes;
	my $ALMType;
	my $currWC;
	my $ALMProject;
	my $wfa_RelatePrimaryChildren;
	my $WC_UK;
	my $ActList;
	my $currAct;
	my %RelatedActList=();
	my $PrimaryActivity;
	my $ActMastership;
	my $IsActRelated;
	my $currRelatedList;
	my $IsActRelatedBack;
	my $currRelatedBackList;
	
	$logger->log_print_message("wfa_Relate_Activities() ͂Ă܂\n");

	# Query ALMTasks that are not completed and have activities
	$querydef = $session->BuildQuery("ALMTask");
	$querydef->BuildField("ID");
	$querydef->BuildField("Project");
	$querydef->BuildField("Type");
	$operator = $querydef->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
	$operator->BuildFilter("State", $CQPerlExt::CQ_COMP_OP_EQ,\@active_open);
	$operator->BuildFilter("Project.Obsolete", $CQPerlExt::CQ_COMP_OP_NEQ,["Yes"]);
	$operator->BuildFilter("Activities", $CQPerlExt::CQ_COMP_OP_IS_NOT_NULL, [""]);
	$resultset = $session->BuildResultSet($querydef);
	$resultset->EnableRecordCount();			
	$resultset->Execute;	
	$count = $resultset->GetRecordCount();	
	while ($resultset->MoveNext eq $CQPerlExt::CQ_SUCCESS) {
		$ALMTask=$resultset->GetColumnValue(1);
		$ALMProject=$resultset->GetColumnValue(2);	
		$ALMType=$resultset->GetColumnValue(3);			
        $ATH{$ALMTask}="$ALMProject" . "ALMTask" . "$ALMType " . "Type";
	}
	# Process for each task
 	foreach my $Task ( keys %ATH ) {
	 	$currTask=$session->GetEntity("ALMTask","$Task");
	 	$WC_UK=$ATH{$Task};
	  	eval{$currWC=$session->GetEntity("ALMWorkConfiguration","$WC_UK")};
 		if ($@){
	 		next;
	 	}
	 	$wfa_RelatePrimaryChildren=$currWC->GetFieldStringValue("wfa_RelatePrimaryChildren");
	 	next if ("$wfa_RelatePrimaryChildren" ne "Yes");
	 	$ActList = $currTask->GetFieldStringValueAsList("Activities");
	 	# Create a list of activities that are related using PrimaryActivity value and are locally mastered
		foreach my $Activity (@$ActList) {
			$currAct=$session->GetEntity("ALMActivity","$Activity");
			$PrimaryActivity=$currAct->GetFieldStringValue("PrimaryActivity");
			$ActMastership=$currAct->GetFieldStringValue("ratl_mastership");
			if (("$ActMastership" eq "$localreplica") && ("$PrimaryActivity" == 1)){
				$RelatedActList{"$Activity"}=1;
			}
		}
		# Now relate the activities to each other if not related already
 		foreach my $ActToModify ( keys %RelatedActList ){
 				$currAct=$session->GetEntity("ALMActivity","$ActToModify");
 				foreach my $ActToAdd ( keys %RelatedActList ){
  						next if ("$ActToModify" eq "$ActToAdd");
  						%attributes = ();
 						$currRelatedList=$currAct->GetFieldStringValueAsList("ActivitiesRelated");
 						$IsActRelated = grep(/($ActToAdd)/, @$currRelatedList);
 						$currRelatedBackList=$currAct->GetFieldStringValueAsList("ActivitiesRelatedBack");
 						$IsActRelatedBack = grep(/($ActToAdd)/, @$currRelatedBackList);
 						if ( (! $IsActRelated ) && (! $IsActRelatedBack)) {
	 						$logger->log_print_message("ANeBreB $ActToAdd  ANeBreB $ActToModify ̊֘AANeBreB ^uɒǉĂ܂\n");
	 						$attributes{"ActivitiesRelated"}=$ActToAdd;
	 						if (AddEntityFieldVals("ALMActivity",$ActToModify,"Modify",%attributes) == 0){
								$logger->log_print_error(__LINE__ ." ANeBreB:$ActToModify ̕ύX̃G[\n");
	 							return 0;	
 							}
 							else{
	 							$logger->log_print_message("ALMActivity:$ActToModify ɕύX܂\n");
 							}
	 					}
 						
 				}
		}
		%RelatedActList=();
 	}
 	$logger->log_print_message("wfa_Relate_Activities() I܂\n");	
	return 1;
}
###################################################################################################################
# sub wfa_Relate_Tasks - Relate primary tasks using the Related tasks tab
###################################################################################################################
sub wfa_Relate_Tasks { 
	
	my $querydef;
	my $operator;
	my $resultset;
	my $count;
	my $ALMRequest;
	my %ARH=();
	my $failure;
	my $currRequest;
	my %attributes;
	my $ALMType;
	my $currWC;
	my $ALMProject;
	my $wfa_RelatePrimaryChildren;
	my $WC_UK;
	my $TaskList;
	my $currTask;
	my %RelatedTaskList=();
	my $PrimaryTask;
	my $TaskMastership;
	my $IsTaskRelated;
	my $currRelatedList;
	my $IsTaskRelatedBack;
	my $currRelatedBackList;
	my $TaskPriority;
	
	$logger->log_print_message("wfa_Relate_Tasks() ͂Ă܂\n");

	# Query ALMRequests that are not completed and have tasks
	$querydef = $session->BuildQuery("ALMRequest");
	$querydef->BuildField("ID");
	$querydef->BuildField("Project");
	$querydef->BuildField("Type");
	$operator = $querydef->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
	$operator->BuildFilter("State", $CQPerlExt::CQ_COMP_OP_EQ,["Opened"]);
	$operator->BuildFilter("Project.Obsolete", $CQPerlExt::CQ_COMP_OP_NEQ,["Yes"]);
	$operator->BuildFilter("Tasks", $CQPerlExt::CQ_COMP_OP_IS_NOT_NULL, [""]);
	$resultset = $session->BuildResultSet($querydef);
	$resultset->EnableRecordCount();			
	$resultset->Execute;	
	$count = $resultset->GetRecordCount();	
	while ($resultset->MoveNext eq $CQPerlExt::CQ_SUCCESS) {
		$ALMRequest=$resultset->GetColumnValue(1);
		$ALMProject=$resultset->GetColumnValue(2);	
		$ALMType=$resultset->GetColumnValue(3);			
        $ARH{$ALMRequest}="$ALMProject" . "ALMRequest" . "$ALMType " . "Type";
	}
	# Process for each request
 	foreach my $Request ( keys %ARH ) {
	 	$currRequest=$session->GetEntity("ALMRequest","$Request");
	 	$WC_UK=$ARH{$Request};
	  	eval{$currWC=$session->GetEntity("ALMWorkConfiguration","$WC_UK")};
 		if ($@){
	 		next;
	 	}
	 	$wfa_RelatePrimaryChildren=$currWC->GetFieldStringValue("wfa_RelatePrimaryChildren");
	 	next if ("$wfa_RelatePrimaryChildren" ne "Yes");
	 	$TaskList = $currRequest->GetFieldStringValueAsList("Tasks");
	 	# Create a list of tasks that are related using PrimaryTask value and are locally mastered
		foreach my $Task (@$TaskList) {
			$currTask=$session->GetEntity("ALMTask","$Task");
			$PrimaryTask=$currTask->GetFieldStringValue("PrimaryTask");
			$TaskPriority=$currTask->GetFieldStringValue("Priority");
			$TaskMastership=$currTask->GetFieldStringValue("ratl_mastership");
			if (("$TaskMastership" eq "$localreplica") && ("$PrimaryTask" == 1) && ("$TaskPriority" ne "")){
				$RelatedTaskList{"$Task"}=1;
			}
		}
		# Now relate the tasks to each other if not related already
 		foreach my $TaskToModify ( keys %RelatedTaskList ){
 				$currTask=$session->GetEntity("ALMTask","$TaskToModify");
 				foreach my $TaskToAdd ( keys %RelatedTaskList ){
  						next if ("$TaskToModify" eq "$TaskToAdd");
  						%attributes=();
 						$currRelatedList=$currTask->GetFieldStringValueAsList("TasksRelated");
 						$IsTaskRelated = grep(/($TaskToAdd)/, @$currRelatedList);
 						$currRelatedBackList=$currTask->GetFieldStringValueAsList("TasksRelatedBack");
 						$IsTaskRelatedBack = grep(/($TaskToAdd)/, @$currRelatedBackList);
 						if ((! $IsTaskRelated ) && (! $IsTaskRelatedBack )){
	 						$logger->log_print_message("^XN $TaskToAdd ^XN $TaskToModify ̊֘A^XN ^uɒǉĂ܂\n");
	 						$attributes{"TasksRelated"}=$TaskToAdd;
	 						if (AddEntityFieldVals("ALMTask",$TaskToModify,"Modify",%attributes) == 0){
								$logger->log_print_error(__LINE__ ." ^XN $TaskToModify ̕ύX̃G[\n");
	 							return 0;	
 							}
 							else{
	 							$logger->log_print_message("ALMTask:$TaskToModify ɕύX܂\n");
 							}
	 					}
 						
 				}
		}
		%RelatedTaskList=();
 	}
 	$logger->log_print_message("wfa_Relate_Tasks() I܂\n");	
	return 1;
}
###################################################################################################################
# sub wfa_Sequence_Activities - Move Successors to a target state when the Predecessors moves to a target state
###################################################################################################################
sub wfa_Sequence_Activities { 
	
	my $querydef;
	my $operator;
	my $resultset;
	my @active_open=("Activated","Opened");
	my $currAct;
	my $count;
	my $ALMTask;
	my %ATH=();
	my $failure;
	my $currTask;
	my %attributes;
	my $ALMType;
	my $currWC;
	my $ALMProject;
	my $WC_UK;
	my $ActList;
	my $ActMastership;
	my $successorsList;
	my $predecessorsList;
	my $predState;
	my $succState;
	my $SelectionFlag;
	my $wfa_rescode;
	my $ActType;
	my $ActState;
	my %succHash=();
	my %predHash=();
	my $Type;
	my $predTypeMatched;
	my $succTypeMatched;
	my $roleRecs;
	my $Owner;
	my $roleEntity;
	my $wfa_rescodes;
	my $ActResCode;
	my $predResCodeMatched;
	my $wfa_rescodes;

	
	$logger->log_print_message("wfa_Sequence_Activities() ͂Ă܂\n");

	# Query ALMTasks that are not completed and have activities
	$querydef = $session->BuildQuery("ALMTask");
	$querydef->BuildField("ID");
	$querydef->BuildField("Project");
	$querydef->BuildField("Type");
	$operator = $querydef->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
	$operator->BuildFilter("State", $CQPerlExt::CQ_COMP_OP_EQ,\@active_open);
	$operator->BuildFilter("Project.Obsolete", $CQPerlExt::CQ_COMP_OP_NEQ,["Yes"]);
	$operator->BuildFilter("Activities", $CQPerlExt::CQ_COMP_OP_IS_NOT_NULL, [""]);
	$resultset = $session->BuildResultSet($querydef);
	$resultset->EnableRecordCount();			
	$resultset->Execute;	
	$count = $resultset->GetRecordCount();	
	while ($resultset->MoveNext eq $CQPerlExt::CQ_SUCCESS) {
		$ALMTask=$resultset->GetColumnValue(1);
		$ALMProject=$resultset->GetColumnValue(2);	
		$ALMType=$resultset->GetColumnValue(3);			
        $ATH{$ALMTask}="$ALMProject" . "ALMTask" . "$ALMType " . "Type";
	}
	# Find the children activities and verify if they satisfy the pred/succ criteria and perform action
 	TASK: foreach my $Task ( keys %ATH ) {
	 	$currTask=$session->GetEntity("ALMTask","$Task");
	 	# Obtain the policy and store the policy settings
	 	$WC_UK=$ATH{$Task};
	  	eval{$currWC=$session->GetEntity("ALMWorkConfiguration","$WC_UK")};
 		next if ($@);
	 	$predecessorsList=$currWC->GetFieldStringValueAsList("wfa_Predecessors");
	 	$successorsList=$currWC->GetFieldStringValueAsList("wfa_Successors");
	 	$SelectionFlag=$currWC->GetFieldStringValue("wfa_flag");
	 	next if (($$predecessorsList[0] eq "") ||  ($$successorsList[0] eq "") || ($SelectionFlag eq "none"));
	 	if ($SelectionFlag eq "Open"){
		 	$predState=$currWC->GetFieldStringValue("wfa_pState_for_open");
		 	$succState="Opened";
		 }
	 	if ($SelectionFlag eq "Activate"){
		 	$predState=$currWC->GetFieldStringValue("wfa_pState_for_activate");
		 	$succState="Activated";
		 }
		if ($SelectionFlag eq "Complete"){
		 	$predState=$currWC->GetFieldStringValue("wfa_pState_for_complete");
		 	$succState="Completed";
		 	$wfa_rescode=$currWC->GetFieldStringValue("wfa_ResolutionCode");
		 	$wfa_rescodes=$currWC->GetFieldStringValueAsList("wfa_ResolutionCodes");
		}
	 	%succHash=();
	    %predHash=();
	    # Initialize the hash of the predecessor types
		foreach (@$predecessorsList) {$predHash{$_}=0;}
		
	 	$ActList = $currTask->GetFieldStringValueAsList("Activities");
	 	# There has to be at least one activity with a type that matches the type in the predecessor list
 		foreach my $Activity (@$ActList) {
 			$currAct=$session->GetEntity("ALMActivity","$Activity");
 			$ActState=$currAct->GetFieldStringValue("State");
 			$ActMastership=$currAct->GetFieldStringValue("ratl_mastership");
 			$ActType=$currAct->GetFieldStringValue("Type");
 			$ActResCode=$currAct->GetFieldStringValue("ResolutionCode");
			$predTypeMatched= grep(/($ActType)/, @$predecessorsList);
			$predResCodeMatched=grep(/($ActResCode)/, @$wfa_rescodes);
			if (($predTypeMatched) && ($ActState eq $predState)){
				if (($SelectionFlag ne "Complete") || 
				   (($SelectionFlag eq "Complete") && ($predState ne "Completed")) ||
				   (($predState eq "Completed") && ($predResCodeMatched) && ($SelectionFlag eq "Complete")) ||
				   (($predState eq "Completed") && ($$wfa_rescodes[0] eq "") && ($SelectionFlag eq "Complete"))){
					$predHash{"$ActType"}++;
				}
			}
			$succTypeMatched= grep(/($ActType)/, @$successorsList);
			# Select the activity if locally mastered, has one of the types in the successor list, and is not already in the target state
			if (($succTypeMatched) && ($ActMastership eq $localreplica) && ($ActState ne $succState)){
				$succHash{$Activity}=$ActState;
			}
		}
		# If any of the pred/succ hashes are blank then go to the next task
		next if ( !(keys %predHash) && !(keys %succHash));
		# If the pred hash contains types that are not matched then go the next task
		foreach my $anyMatches ( values %predHash ) {
			next TASK if ($anyMatches == 0);
 		}
 		# For the remaining activities, move to the target state, set the owner and res code
		foreach my $succAct ( keys %succHash ) {
			$Owner="";
			%attributes=();
	 		$currAct = $session->GetEntity("ALMActivity", $succAct);
  			$roleRecs = $currAct->GetFieldStringValueAsList("Roles");
			if ($$roleRecs[0] ne ""){
				@{$roleRecs}= sort(@{$roleRecs});
	 	    	$roleEntity = $session->GetEntity("ALMRole",$$roleRecs[0]);
				$Owner= $roleEntity->GetFieldStringValue("Primary");
				if ($Owner){
					%attributes = (
						Owner => $Owner,
	 				);
 				}
			}
			if ($Owner eq ""){
				$logger->log_print_error(__LINE__ ." ALMActivity:$succAct ̓KL҂邱Ƃł܂ł");
				return 0;	
			}
			if (($SelectionFlag eq "Open") && ($succHash{$succAct} eq "Submitted")){
				$logger->log_print_message("ALMActivity:$succAct JԂɈړĂ܂\n");
				$logger->log_print_message("ALMActivity:$succAct Ɋɐݒ肳ĂȂꍇAL҂ $Owner ɐݒ肳܂\n");
				if (SetEntityFieldVals("ALMActivity",$succAct, "Open", %attributes) == 0){
					$logger->log_print_error(__LINE__ ." ANeBreB:$succAct ̕ύX̃G[\n");
					return 0;	
				}
				else{
					$logger->log_print_message("ALMActivity:$succAct ɕύX܂\n");
				}
			}
			if (($SelectionFlag eq "Activate") && (($succHash{$succAct} eq "Submitted") || ($succHash{$succAct} eq "Opened"))){
				$logger->log_print_message("ALMActivity:$succAct ANeBuԂɈړĂ܂\n");
				if (SetEntityFieldVals("ALMActivity",$succAct, "Activate", %attributes) == 0){
					$logger->log_print_error(__LINE__ ." ANeBreB:$succAct ̕ύX̃G[\n");
					$logger->log_print_message("ALMActivity:$succAct Ɋɐݒ肳ĂȂꍇAL҂ $Owner ɐݒ肳܂\n");
					return 0;	
				}
				else{
					$logger->log_print_message("ALMActivity:$succAct ɕύX܂\n");
				}
			}
			if (($SelectionFlag eq "Complete") && (($succHash{$succAct} eq "Opened") || ($succHash{$succAct} eq "Activated"))){
				$logger->log_print_message("ALMActivity:$succAct ԂɈړĂ܂\n");
				$attributes{"ResolutionCode"}=$wfa_rescode;
				$attributes{"ResolutionSummary"}=$wfa_rescode;
				if (SetEntityFieldVals("ALMActivity",$succAct, "Complete", %attributes) == 0){
					$logger->log_print_error(__LINE__ ." ANeBreB:$succAct ̕ύX̃G[\n");
					return 0;	
				}
				else{
					$logger->log_print_message("ALMActivity:$succAct ɕύX܂\n");
				}
			}
 		}
  	}
  	$logger->log_print_message("wfa_Sequence_Activities() I܂\n");	
	return 1;
}
###################################################################################################################
# sub wfa_Sequence_Tasks - Move Successors to a target state when the Predecessors moves to a target state
###################################################################################################################
sub wfa_Sequence_Tasks{ 
	
	my $querydef;
	my $operator;
	my $resultset;
	my $currTask;
	my $count;
	my $ALMRequest;
	my %ARH=();
	my $failure;
	my $currRequest;
	my %attributes;
	my $ALMType;
	my $currWC;
	my $ALMProject;
	my $WC_UK;
	my $TaskList;
	my $TaskMastership;
	my $successorsList;
	my $predecessorsList;
	my $predState;
	my $succState;
	my $SelectionFlag;
	my $wfa_rescode;
	my $TaskType;
	my $TaskState;
	my %succHash=();
	my %predHash=();
	my $Type;
	my $predTypeMatched;
	my $succTypeMatched;
	my $roleRecs;
	my $Owner;
	my $roleEntity;
	my $wfa_rescodes;
	my $TaskResCode;
	my $predResCodeMatched;
	my $wfa_rescodes;
	my $primaryOwner;

	
	$logger->log_print_message("wfa_Sequence_Tasks() ͂Ă܂\n");

	# Query ALMRequests that are not completed and have Tasks
	$querydef = $session->BuildQuery("ALMRequest");
	$querydef->BuildField("ID");
	$querydef->BuildField("Project");
	$querydef->BuildField("Type");
	$operator = $querydef->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
	$operator->BuildFilter("State", $CQPerlExt::CQ_COMP_OP_EQ,["Opened"]);
	$operator->BuildFilter("Project.Obsolete", $CQPerlExt::CQ_COMP_OP_NEQ,["Yes"]);
	$operator->BuildFilter("Tasks", $CQPerlExt::CQ_COMP_OP_IS_NOT_NULL, [""]);
	$resultset = $session->BuildResultSet($querydef);
	$resultset->EnableRecordCount();			
	$resultset->Execute;	
	$count = $resultset->GetRecordCount();	
	while ($resultset->MoveNext eq $CQPerlExt::CQ_SUCCESS) {
		$ALMRequest=$resultset->GetColumnValue(1);
		$ALMProject=$resultset->GetColumnValue(2);	
		$ALMType=$resultset->GetColumnValue(3);			
        $ARH{$ALMRequest}="$ALMProject" . "ALMRequest" . "$ALMType " . "Type";
	}
	# Find the children tasks and verify if they satisfy the pred/succ criteria and perform action
 	REQUEST: foreach my $Request ( keys %ARH ) {
	 	$currRequest=$session->GetEntity("ALMRequest","$Request");
	 	# Obtain the policy and store the policy settings
	 	$WC_UK=$ARH{$Request};
	  	eval{$currWC=$session->GetEntity("ALMWorkConfiguration","$WC_UK")};
 		next if ($@);
	 	$predecessorsList=$currWC->GetFieldStringValueAsList("wfa_Predecessors");
	 	$successorsList=$currWC->GetFieldStringValueAsList("wfa_Successors");
	 	$SelectionFlag=$currWC->GetFieldStringValue("wfa_flag");
	 	next if (($$predecessorsList[0] eq "") ||  ($$successorsList[0] eq "") || ($SelectionFlag eq "none"));
	 	if ($SelectionFlag eq "Open"){
		 	$predState=$currWC->GetFieldStringValue("wfa_pState_for_open");
		 	$succState="Opened";
		}
	 	if ($SelectionFlag eq "Activate"){
		 	$predState=$currWC->GetFieldStringValue("wfa_pState_for_activate");
		 	$succState="Activated";
		}
		if ($SelectionFlag eq "Complete"){
		 	$predState=$currWC->GetFieldStringValue("wfa_pState_for_complete");
		 	$succState="Completed";
		 	$wfa_rescode=$currWC->GetFieldStringValue("wfa_ResolutionCode");
		 	$wfa_rescodes=$currWC->GetFieldStringValueAsList("wfa_ResolutionCodes");
		}
	 	%succHash=();
	    %predHash=();
	     # Initialize the hash of the predecessor types
		foreach (@$predecessorsList) {$predHash{$_}=0;}
		
	 	$TaskList = $currRequest->GetFieldStringValueAsList("Tasks");
	 	# There has to be at least one task with a type that matches the type in the predecessor list
 		foreach my $Task (@$TaskList) {
 			$currTask=$session->GetEntity("ALMTask","$Task");
 			$TaskState=$currTask->GetFieldStringValue("State");
 			$TaskMastership=$currTask->GetFieldStringValue("ratl_mastership");
 			$TaskType=$currTask->GetFieldStringValue("Type");
 			$TaskResCode=$currTask->GetFieldStringValue("ResolutionCode");
			$predTypeMatched= grep(/($TaskType)/, @$predecessorsList);
			$predResCodeMatched=grep(/($TaskResCode)/, @$wfa_rescodes);
			if (($predTypeMatched) && ($TaskState eq $predState)){
				if (($SelectionFlag ne "Complete") || 
				   (($SelectionFlag eq "Complete") && ($predState ne "Completed")) ||
				   (($predState eq "Completed") && ($predResCodeMatched) && ($SelectionFlag eq "Complete")) ||
				   (($predState eq "Completed") && ($$wfa_rescodes[0] eq "") && ($SelectionFlag eq "Complete"))){
					$predHash{"$TaskType"}++;
				}
			}
			# Select the task if locally mastered, has one of the types in the successor list, and is not already in the target state
			$succTypeMatched= grep(/($TaskType)/, @$successorsList);
			if (($succTypeMatched) && ($TaskMastership eq $localreplica) && ($TaskState ne $succState)){
				$succHash{$Task}=$TaskState;
			}
		}
		# If any of the pred/succ hashes are blank then go to the next request
		next if ( !(keys %predHash) && !(keys %succHash));
		foreach my $anyMatches ( values %predHash ) {
			next REQUEST if ($anyMatches == 0);
 		}
 		# For the remaining tasks, move to the target state, set the owner and res code
		foreach my $succTask ( keys %succHash ) {
			$Owner="";
			%attributes=();
	 		$primaryOwner = $currTask->GetFieldStringValue("PrimaryRole.Primary");
 			$roleRecs = $currTask->GetFieldStringValueAsList("Roles");
 			if ($primaryOwner){
 				$Owner=$primaryOwner;		
 			}
 			elsif ($$roleRecs[0] ne ""){
				@{$roleRecs}= sort(@{$roleRecs});
	 	    	$roleEntity = $session->GetEntity("ALMRole",$$roleRecs[0]);
				$Owner= $roleEntity->GetFieldStringValue("Primary");

			}
			else {
				$logger->log_print_error(__LINE__ ." ALMTask:$succTask ̓KL҂邱Ƃł܂ł");
				return 0;	
			}
			%attributes = (
				Owner => $Owner,
	 		);
			if ($SelectionFlag eq "Open"){
				$logger->log_print_message("[Nt[̎|V[: 㑱^XÑI[v^XN $succTask ɖ܂\n");
			}
			if (($SelectionFlag eq "Activate") && ($succHash{$succTask} eq "Opened")){
				$logger->log_print_message("ALMTask:$succTask ANeBuԂɈړĂ܂\n");
				if (SetEntityFieldVals("ALMTask",$succTask, "Activate", %attributes) == 0){
					$logger->log_print_error(__LINE__ ." ^XN $succTask ̕ύX̃G[\n");
					$logger->log_print_message("ALMTask:$succTask Ɋɐݒ肳ĂȂꍇAL҂ $Owner ɐݒ肳܂\n");
					return 0;	
				}
				else{
					$logger->log_print_message("ALMTask:$succTask ɕύX܂\n");
				}
			}
			if (($SelectionFlag eq "Complete") && (($succHash{$succTask} eq "Opened") || ($succHash{$succTask} eq "Activated"))){
				$logger->log_print_message("ALMTask:$succTask ԂɈړĂ܂\n");
				$attributes{"ResolutionCode"}=$wfa_rescode;
				$attributes{"ResolutionSummary"}=$wfa_rescode;
				if (SetEntityFieldVals("ALMTask",$succTask, "Complete", %attributes) == 0){
					$logger->log_print_error(__LINE__ ." ^XN $succTask ̕ύX̃G[\n");
					return 0;	
				}
				else{
					$logger->log_print_message("ALMTask:$succTask ɕύX܂\n");
				}
			}
 		}
  	}
  	$logger->log_print_message("wfa_Sequence_Tasks() I܂\n");	
	return 1;
}